#default_quotation "expr";
open Camlp4.PreCast;
open Common;
open Ml22d;

value short = ref False;

value fresh_short_name =
  let count = ref (Char.code 'a'-1) in
  fun orig -> do {
    incr count;
    if count.val > Char.code 'z' then orig
    else String.make 1 (Char.chr count.val)
};

class parse =
  object(o)

    method expr =
      fun
      [ << match $e$ with [ ($p1$, $p2$) -> $e1$ ] >> |
        << let ($p1$, $p2$) = $e1$ in $e$ >> ->
          MatchPair (o#expr e1) (o#patt p1) (o#patt p2) (o#expr e)
      | << match $e$ with [ Inl $p1$ -> $e1$ | Inr $p2$ -> $e2$ ] >> |
        << match $e$ with [ Inr $p2$ -> $e2$ | Inl $p1$ -> $e1$ ] >> ->
          MatchCase (o#expr e) (o#patt p1, o#expr e1) (o#patt p2, o#expr e2)
      | << match $e$ with [ True -> $e1$ | False -> $e2$ ] >> |
        << match $e$ with [ False -> $e2$ | True -> $e1$ ] >> |
        << match $e$ with [ Towards -> $e1$ | Away -> $e2$ ] >> |
        << match $e$ with [ Away -> $e2$ | Towards -> $e1$ ] >> ->
          MatchCase (o#expr e) ("_", o#expr e1) ("_", o#expr e2)
      | << $lid:var$ $e$ >> -> Fun1 (o#func_name var) (o#expr e)
      | << $lid:var$ $e$ $e2$ >> -> Fun2 (o#func_name var) (o#expr e) (o#expr e2)
      | << Inl $e$ >> -> Inl (o#expr e)
      | << Inr $e$ >> -> Inr (o#expr e)
      | << True >> -> Inl Unit
      | << False >> -> Inr Unit
      | << None >> -> Inl Unit
      | << Medium >> -> Inr (Inl Unit (*(Var "_")*))
      | << All >> -> Inr (Inr (Inl Unit(*(Var "_")*)))
      | << Error >> -> Unit (* XXX ERROR XXX *)
      | << Towards >> -> Inl Unit
      | << Away >> -> Inr Unit
      | << () >> -> Unit
      | << let $p$ = $e1$ in $e2$ >> -> Let (o#patt p) (o#expr e1) (o#expr e2)
      | << $lid:var$ >> -> Var var
      | << ($e1$, $e2$) >> -> Pair (o#expr e1) (o#expr e2)
      | e -> bad_expr "method expr" e ];

    method patt =
      fun
      [ <:patt< $lid:s$ >> -> s
      | <:patt< _ >> -> "_"
      | <:patt< () >> -> "_"
      | p -> bad_patt "method patt" p ];

    value short_names = do {
      let h = Hashtbl.create 101;
      Hashtbl.add h "main" "main";
      h
    };

    method func_name s =
      if short.val then
        try Hashtbl.find short_names s
        with
        [ Not_found ->
            let n = fresh_short_name s in
            let () = Hashtbl.add short_names s n in n ]
      else s;

    method func st acc =
      let f x = [x :: acc] in
      match st with
      [ <:str_item< value $rec:_$ $lid:var$ $p1$ $p2$ = $e$ >> ->
          f (compile_fun2 (o#func_name var) (o#patt p1) (o#patt p2) (o#expr e))
      | <:str_item< value $rec:_$ $lid:var$ $p$ = $e$ >> ->
          f (compile_fun1 (o#func_name var) (o#patt p) (o#expr e))
      | <:str_item< $st1$; $st2$ >> ->
          o#func st1 (o#func st2 acc)
      | st -> bad_str_item "method func" st ];

    method func_ast st acc =
      let f x = [x :: acc] in
      match st with
      [ <:str_item< value $rec:_$ $lid:var$ $p1$ $p2$ = $e$ >> ->
          f { fname = o#func_name var; body = o#expr e;
               arg1 = o#patt p1; arg2 = Some (o#patt p2) }
      | <:str_item< value $rec:_$ $lid:var$ $p$ = $e$ >> ->
          f { fname = o#func_name var; arg1 = o#patt p;
               arg2 = None; body = o#expr e }
      | <:str_item< $st1$; $st2$ >> ->
          o#func_ast st1 (o#func_ast st2 acc)
      | st -> bad_str_item "method func" st ];
end;

value of_modules str_item = (new parse)#func str_item [];

value ast_of_modules str_item = (new parse)#func_ast str_item [];

Camlp4.Options.add "-short" (Arg.Set short) "Use short names for functions";
