open Big_int
open ExtList

type nb_proofs = 
  | Zero
  | Finite of big_int (* Not supposed to be zero *)
  | Infinite

let to_string_nb_proofs = function
  | Zero -> "0"
  | Finite i -> Printf.sprintf "%s" (string_of_big_int i)
  | Infinite -> 
      if !Options.result = Options.LaTeX then
        "\\infty"
      else
        "inf"

let ( +% ) x y = match x, y with
  | Infinite, _ | _, Infinite -> Infinite
  | Finite x, Zero | Zero , Finite x -> Finite x
  | Finite x, Finite y -> Finite (add_big_int x y)
  | Zero, Zero -> Zero

let ( *% ) x y = match x, y with
  | Zero, _ | _, Zero -> Zero
  | Infinite, _ | _, Infinite -> Infinite
  | Finite x, Finite y -> Finite (mult_big_int x y)

let ( >=% ) x y = match x, y with
  | Infinite, _ -> true
  | Zero, Zero -> true
  | Zero, _ -> false
  | Finite i1, Finite i2 -> ge_big_int i1 i2
  | Finite _, Infinite -> false
  | Finite _, Zero -> true

let max_nb_proofs x y = match x, y with
  | Infinite, _ | _, Infinite -> Infinite
  | Finite i1, Finite i2 -> Finite (max_big_int i1 i2)
  | Finite i, Zero | Zero, Finite i -> Finite i
  | Zero, Zero -> Zero


module CountState (Pb: Set.OrderedType) =
struct

module ProdPbs = Set.Make(Pb)
module SumProdPbs = Set.Make(ProdPbs)

type pb_set = SumProdPbs.t
let no_loop = SumProdPbs.empty


let is_looping s pb = 
  SumProdPbs.exists (ProdPbs.mem pb) s

let contains_singleton s pb =
  let eq_singleton = ((=) (ProdPbs.singleton pb)) in 
    SumProdPbs.exists eq_singleton s

let singleton pb =
  SumProdPbs.singleton (ProdPbs.singleton pb)

let sum_of_list_prods =
  List.fold_left (fun s p -> SumProdPbs.add p s) no_loop

let remove_loop s pb = 
  let s' = SumProdPbs.elements s in
  let s'' = List.map (ProdPbs.remove pb) s' in
    SumProdPbs.remove ProdPbs.empty (sum_of_list_prods s'')

(* let remove_loop s pb = 
  SumProdPbs.filter (fun p -> not (ProdPbs.mem pb p)) s *)


let product s1 s2 =
  let l1 = SumProdPbs.elements s1
  and l2 = SumProdPbs.elements s2 in
  let l = List.cartesian l1 l2 in
  let l' = List.map (fun (p1, p2) -> ProdPbs.union p1 p2) l in
    sum_of_list_prods l'

type count_state = {
  nb : nb_proofs ;
  explored : bool ;
  loop : pb_set }

let unprovable = {
  nb = Zero ;
  explored = true ;
  loop = no_loop }

let is_unprovable = (=) unprovable

let ( +$ ) x y = { 
  nb = x.nb +% y.nb ;
  explored = x.explored && y.explored ;
  loop = SumProdPbs.union x.loop y.loop }

let ( *$ ) x y =
 if x = unprovable || y = unprovable then unprovable
 else 
   { nb = x.nb *% y.nb ; explored = x.explored && y.explored ;
     loop = SumProdPbs.union (product x.loop y.loop) 
              (SumProdPbs.union 
                 (if x.nb <> Zero then y.loop else no_loop)
                 (if y.nb <> Zero then x.loop else no_loop)) }

let count_one = {
  nb = Finite unit_big_int;
  loop = no_loop;
  explored = true }

let count_zero = {
  nb = Zero;
  loop = no_loop;
  explored = true }

let count_infinity = {
  nb = Infinite;
  loop = no_loop;
  explored = true }

end
