type prop = 
  | PVar of int 
  | Arrow of prop * prop 
  | And of prop list 
  | Or of int * prop list


let lpb = [

(* |- t *) 
(* 0 *)
  [], PVar 1;

(* t |- t *) 
(* 1 *)
  [PVar 1], PVar 1;

(* t2 -> t1, t2 |- t1 *)
(* 1 *)
  [Arrow (PVar 2, PVar 1) ; PVar 2] , PVar 1;

(* t1 -> t1, t1 |- t1 *)
(* inf *)
  [Arrow (PVar 1, PVar 1) ; PVar 1] , PVar 1;

(* |- t1 -> (t1 -> t2) -> t2 *)
(* 1 *)
  [], Arrow (PVar 1, Arrow (Arrow (PVar 1, PVar 2), PVar 2));

(* t1 -> t2 -> (t1 -> t2 -> t3) -> t3 *)
(* 1 *)
  [], Arrow (PVar 1, 
      Arrow (PVar 2, 
      Arrow (Arrow (PVar 1, 
             Arrow (PVar 2, PVar 3)), 
             PVar 3)));

(* t1, t2, t1 -> t2 -> t3 |- t3 *)
(* 1 *)
  [PVar 1; PVar 2 ; Arrow (PVar 1, Arrow (PVar 2, PVar 3))], PVar 3;

(* |- (t1 * t1) -> (t1 * t1) *)
(* 4 *)
  [], Arrow (And [PVar 1 ; PVar 1], And [PVar 1 ; PVar 1]);

(* t1 -> (t2 * t3), t1 |- t2 * t3 *)
(* 1 *)
  [Arrow (PVar 1, And [PVar 2 ; PVar 3]) ; PVar 1],
  And [PVar 2 ; PVar 3];

(* t1 -> (t2 * t3), t1 -> (t3 * t2), t1 -> t2, t1 |- t2 * t3 *)
(* 6 *)
  [Arrow (PVar 1, And [PVar 2 ; PVar 3]) ; 
   Arrow (PVar 1, And [PVar 3 ; PVar 2]) ;
   Arrow (PVar 1, PVar 2) ; 
   PVar 1],
  And [PVar 2 ; PVar 3];

(* t1 -> t2, t2 -> t1, t1 -> t3, t2 -> t3, t1, t2 |- t3 *)
(* inf *)
  [Arrow (PVar 1, PVar 2);
   Arrow (PVar 2, PVar 1);
   Arrow (PVar 1, PVar 3);
   Arrow (PVar 2, PVar 3);
   PVar 1 ; PVar 2],
  PVar 3;

(* t1 -> t1, t1, t2 -> t2, t2 |- t1 * t2 *)
(* inf *)
  [Arrow (PVar 1, PVar 1); PVar 1;
   Arrow (PVar 2, PVar 2); PVar 2],
  And [PVar 1 ; PVar 2];

(* (t1 -> t1) -> t1, (t2 -> t2) -> t2 |- t1 * t2 *)
(* inf *)
  [Arrow (Arrow (PVar 1, PVar 1), PVar 1);
   Arrow (Arrow (PVar 2, PVar 2), PVar 2)],
  And [PVar 1 ; PVar 2];

(* t1 * t1 -> t1 |- t1 *)
(* 0 *)
  [Arrow (And [PVar 1 ; PVar 1], PVar 1)], PVar 1;

(* (t1 -> t1) -> t1 |- t1 *)
(* inf *)
  [Arrow (Arrow (PVar 1, PVar 1), PVar 1)], PVar 1;

(* (t1 -> t1) -> t1 |- t1 -> t1 *)
(* inf *)
  [Arrow (Arrow (PVar 1, PVar 1), PVar 1)], (Arrow (PVar 1, PVar 1));

(* Exponential *)
(* 1 *)
  [Arrow (PVar 1, And [PVar 2 ; PVar 3]);
   Arrow (And [PVar 3; PVar 2], And [PVar 4; PVar 5; PVar 6; PVar 7]);
   Arrow (And [PVar 7; PVar 6; PVar 5; PVar 4],
          And [PVar 8;  PVar 9;  PVar 10; PVar 11; 
               PVar 12; PVar 13; PVar 14; PVar 15]);
   PVar 1],
  And [PVar 15; PVar 14; PVar 13; PVar 12;
       PVar 11; PVar 10; PVar  9; PVar 8];

(* t1, t1 -> t2, (t2 * t2) -> (t3 * t4) |- t3 * t4 *)
(* 1 *)
  [PVar 1; Arrow (PVar 1, PVar 2);
   Arrow (And [PVar 2; PVar 2], And [PVar 3; PVar 4])],
  And [PVar 4; PVar 3];


(* t1 \/ t2 |- t1 \/ t2 *)
(* inf (cough) *)
  [Or (1, [PVar 1 ; PVar 2])], Or (1, [PVar 1 ; PVar 2]);

(* t1, t2 |- t1 \/ t2 *)
(* 2 *)
  [PVar 1; PVar 2], Or (1, [PVar 1 ; PVar 2]) ;

(* t2 -> t3 -> t3, t3, t3, t2 -> t3 -> t2, t3 -> t3 -> t1, t1 -> t2 -> t1 |- t1 *)
(* 1 *)
  [Arrow (PVar 2, Arrow (PVar 3, PVar 3)) ; PVar 3 ; PVar 3 ;
   Arrow (PVar 2, Arrow (PVar 3, PVar 2)) ;
   Arrow (PVar 3, Arrow (PVar 3, PVar 1)) ;
   Arrow (PVar 1, Arrow (PVar 2, PVar 1))], PVar 1
]

let rec prop_to_propt = function
  | PVar i -> Prop.pvar i
  | Arrow (a, b) -> Prop.parrow (prop_to_propt a) (prop_to_propt b) 
  | And l -> Prop.pand (List.map prop_to_propt l)
  | Or (i, l) -> Prop.por i (List.map prop_to_propt l)

open Main

let map_option f = function
  | None -> None
  | Some v -> Some (f v)

let convert_pb_set pb = 
  let env, stoup, goal = pb in 
    { LJTSet.env = Env.EnvSet.of_list (List.map prop_to_propt env) ;
      LJTSet.stoup = map_option prop_to_propt stoup;
      LJTSet.goal = prop_to_propt goal }

let convert_pb_mset pb = 
  let env, stoup, goal = pb in 
    { LJTMSet.env = Env.EnvMultiSet.of_list (List.map prop_to_propt env) ;
      LJTMSet.stoup = map_option prop_to_propt stoup;
      LJTMSet.goal = prop_to_propt goal }
