type t = int

let compare (v1 : t) (v2 : t) = v1 - v2

type path =
  | PAx
  | PNoChoice
  | PCont of (Prop.t * int) * t
  | PImpL of t * t
  | PImpR of t
  | PAndL of (int * int) * t
  | PAndR of t list
  | POrL of Prop.or_nb * t list
  | POrR of Prop.or_nb * (int * int) *  t

let paths : (t, path) Hashtbl.t = Hashtbl.create 57
let ts : (path, t) Hashtbl.t = Hashtbl.create 57

let ax = 0
let nochoice = 1

let _ = 
  Hashtbl.add paths ax PAx;
  Hashtbl.add ts PAx ax;
  Hashtbl.add paths nochoice PNoChoice;
  Hashtbl.add ts PNoChoice nochoice

let next_t : t ref = ref 2

let retrieve_or_create (p : path) =
  try Hashtbl.find ts p
  with Not_found ->
    let t = !next_t in
      Hashtbl.add ts p t;
      Hashtbl.add paths t p;
      incr next_t;
      t

let cont prop i t = retrieve_or_create (PCont ((prop, i), t))
let impl (t1, t2) = retrieve_or_create (PImpL (t1, t2))
let impr t = retrieve_or_create (PImpR t)
let andl i t = retrieve_or_create (PAndL (i, t))
let andr l = retrieve_or_create (PAndR l)
let orl i l = retrieve_or_create (POrL (i, l))
let orr i i' t = retrieve_or_create (POrR (i, i', t))

let path_of_t = Hashtbl.find paths
