open ExtList

type var = int

type t = int

let compare (v1 : t) (v2 : t) = v1 - v2

type prop = 
  | Var of var
  | Arrow of t * t
  | And of t list
  | Or of or_nb * t list
and or_nb = int

let props : (t, prop) Hashtbl.t = Hashtbl.create 57
let ts : (prop, t) Hashtbl.t = Hashtbl.create 57

let next_t : t ref = ref 0

let retrieve_or_create (p : prop) =
  try Hashtbl.find ts p
  with Not_found ->
    let t = !next_t in
      Hashtbl.add ts p t;
      Hashtbl.add props t p;
      incr next_t;
      t

let pvar (v : var) = retrieve_or_create (Var v)
let parrow (p1 : t) (p2 : t) = retrieve_or_create (Arrow (p1, p2))
let pand (l : t list) = retrieve_or_create (And l)
let por i (l : t list) = retrieve_or_create (Or (i, l))

let prop_of_t = Hashtbl.find props

let rec to_string_text t = match prop_of_t t with
  | Var i -> Printf.sprintf "x%d" i
  | Arrow (t1, t2) -> (
      match prop_of_t t1 with
        | Arrow _ -> Printf.sprintf "(%s)->%s" 
            (to_string_text t1) (to_string_text t2)
        | _ -> Printf.sprintf "%s->%s" (to_string_text t1) (to_string_text t2))
  | And l -> Printf.sprintf "(%s)"
      (String.concat " /\\ " (List.map to_string_text l))
  | Or (i, l) -> let sor = Printf.sprintf " \\/%d " i in
      Printf.sprintf "(%s)" (String.concat sor (List.map to_string_text l))

let rec to_string_latex t = match prop_of_t t with
  | Var i -> Printf.sprintf "x_{%d}" i
  | Arrow (t1, t2) -> (
      match prop_of_t t1 with
        | Arrow _ -> Printf.sprintf "(%s) \\rightarrow %s" 
            (to_string_latex t1) (to_string_latex t2)
        | _ -> Printf.sprintf "%s \\rightarrow %s" 
            (to_string_latex t1) (to_string_latex t2))
  | And l -> Printf.sprintf "(%s)"
      (String.concat " \\land " (List.map to_string_latex l))
  | Or (i, l) -> let sor = Printf.sprintf " \\lor_{%d} " i in
      Printf.sprintf "(%s)" (String.concat sor (List.map to_string_latex l))


let to_string_parseable ?(expand_top_or=true) = 
  let rec aux_no_or t = match prop_of_t t with
    | Var i -> Printf.sprintf "%d" i
    | Arrow (t1, t2) -> Printf.sprintf "(ARROW %s %s)"
        (aux_no_or t1) (aux_no_or t2)
    | And l -> Printf.sprintf "(AND %s)" 
        (String.concat " " (List.map aux_no_or l))
    | Or (i, _) -> Printf.sprintf "tor_%d" i
  and aux_or t = match prop_of_t t with
    |Or (i, l) -> Printf.sprintf "[%s]" (String.concat " | " 
                (List.mapi (fun i' p -> Printf.sprintf "`C%d_%d of %s" 
                              i i' (aux_no_or p)) l))
    | _ -> aux_no_or t
  in
    if expand_top_or then aux_or else aux_no_or

let to_string_ocaml ?(expand_top_or=true) = 
  let rec aux_no_or t = match prop_of_t t with
    | Var i -> Printf.sprintf "t%d" i
    | Arrow (t1, t2) -> Printf.sprintf "(%s -> %s)"
        (aux_no_or t1) (aux_no_or t2)
    | And l -> Printf.sprintf "(%s)" 
        (String.concat " * " (List.map aux_no_or l))
    | Or (i, _) -> Printf.sprintf "tor_%d" i
  and aux_or t = match prop_of_t t with
    |Or (i, l) -> Printf.sprintf "[%s]" (String.concat " | " 
                (List.mapi (fun i' p -> Printf.sprintf "`C%d_%d of %s" 
                              i i' (aux_no_or p)) l))
    | _ -> aux_no_or t
  in
    if expand_top_or then aux_or else aux_no_or
        

let to_string p = match !Options.result with
  | Options.LaTeX -> to_string_latex p
  | Options.Text -> to_string_text p


