open Instruction
open State

module State = State

exception Halted

let i64 b = 
  let b = Int64.of_int32 b in
  if Int64.compare b 0L < 0 then Int64.add b (Int64.shift_left 1L 32)
  else b

let division b c =
  Int64.to_int32 (Int64.div (i64 b) (i64 c))

let is_input = ref false

(*** eval *)
let eval s = function
  | ConditionalMove (a,b,c) ->
      if (get_reg s c <> 0l) then set_reg s a (get_reg s b)
  | ArrayIndex (a,b,c) ->
      set_reg s a 
	(Block.get (get_array s (get_reg s b))
	   (Int32.to_int (get_reg s c)))
  | ArrayAmendment (a,b,c) ->
      Block.set (get_array s (get_reg s a))
	(Int32.to_int (get_reg s b))
	(get_reg s c)
  | Addition (a,b,c) ->
      set_reg s a (Int32.add (get_reg s b) (get_reg s c))
  | Multiplication (a,b,c) ->
      set_reg s a (Int32.mul (get_reg s b) (get_reg s c))
  | Division (a,b,c) ->
      set_reg s a (division (get_reg s b) (get_reg s c))
  | NotAnd (a,b,c) ->
      set_reg s a (Int32.lognot (Int32.logand (get_reg s b) (get_reg s c)))
  | Halt ->
      raise Halted
  | Allocation (b,c) ->
      set_reg s b (alloc_array s (get_reg s c))
  | Abandonment c ->
      free_array s (get_reg s c)
  | Output c ->
      print_char (Char.chr (Int32.to_int (get_reg s c)));
      flush stdout
  | Input c ->
      is_input := true;
      set_reg s c
        (try Int32.of_int (Char.code (input_char stdin))
         with End_of_file -> -1l);
      is_input := false
  | LoadProgram (b,c) ->
      load s (get_reg s b) (get_reg s c)
  | Orthography (a,n) ->
      set_reg s a n
(* ***)
      
exception Break
let loop s =
  Printf.eprintf "Using unfunctorized version: %s.\n%!" (State.id);
(*  let break = ref false in *)
  Sys.catch_break true;
(*  Sys.set_signal Sys.sigkill (Sys.Signal_handle 
                                (fun _ -> 
                                   prerr_endline "BREAK !!\n";
                                   flush stderr;
                                   break := true;
                                   if !is_input then raise Break)); *)
  let counter = ref 0 in
(*  let nbsave = ref 1 in *)
(*
  let save () = 
    Printf.eprintf "save to state%i\n" !nbsave;
    State.save_state (Printf.sprintf "state%i" !nbsave) s;
    incr nbsave;
(*    counter := 0; *)
    if !break then exit 0
  in
*)

  Printf.eprintf "Starting evaluation...\n%!";
  try
    while true do
      incr counter;
      if !counter = 1024 * 1024 * 128 then (incr Date.date; counter := 0;
				    State.gc s);
      (*if !break || (!counter = 1024 * 1024 * 128) then save();*)

      let i = fetch_instr s in
      let d = Instruction.of_int32 i in
      (* Format.eprintf "%a@." Instruction.pretty_print d; *)
      eval s d
    done
  with
  | Sys.Break ->
(*      Printf.eprintf "\n*** BREAK received, saving.\n%!";
      save ()  *)
      ()
