type t = int32
type reg = int
type block = Block.t

type state = {
  regs: t array;
  mutable arrays: block array array;
  mutable flist: int;  (* the first available block *)
  mutable pc: int;
}

let delta = 16
let chunk_size = 1 lsl delta
let mask = chunk_size - 1

(*
module Array = struct
  include Array
  let get a i =
    try get a i
    with _ -> failwith (Printf.sprintf "get i=%i size=%i" i (Array.length a))
  let make i x =
    try make i x
    with _ -> failwith (Printf.sprintf "make n=%i" i)
end
*)
 
let get_reg s i = s.regs.(i)
let set_reg s i x = s.regs.(i) <- x
let get_array s i = 
  let i = Int32.to_int i in
  s.arrays.(i lsr delta).(i land mask)

let alloc_array s n =
  let n = Int32.to_int n in
  let i = s.flist in

  let chunk = i lsr delta in
  if chunk >= Array.length s.arrays then (
    let nsize = Array.length s.arrays * 2 + 1 in 
    let a = s.arrays in
    s.arrays <- Array.create nsize [||];
    Array.blit a 0 s.arrays 0 (Array.length a);
  );
  let a = 
    match s.arrays.(chunk) with
      | [||] -> 
	  let a = Array.create chunk_size Block.empty in
	  s.arrays.(chunk) <- a;
	  a
      | a -> a in
  s.flist <- begin
    let k = a.(i land mask) in
    if k == Block.empty then succ i
    else Int32.to_int (Block.get k 0)
  end;
  a.(i land mask) <- Block.create n;
  Int32.of_int i

let free_array s i =
  let i = Int32.to_int i in
  let k = Block.create 1 in
  Block.set k 0 (Int32.of_int s.flist);
  s.arrays.(i lsr delta).(i land mask) <- k;
  s.flist <- i
  
let input_byte accu ic = 
  Int32.logor 
    (Int32.shift_left accu 8)
    (Int32.of_int (Char.code (input_char ic)))

let init ic =
  let n = in_channel_length ic in
  assert(n mod 4 = 0);
  let n = n / 4 in
  let code = Block.create n in
  for i = 0 to n - 1 do
    let a = input_byte 0l ic in
    let a = input_byte a ic in
    let a = input_byte a ic in
    let a = input_byte a ic in
    Block.set code i a
  done;
  let array0 = Array.create chunk_size Block.empty in
  array0.(0) <- code;
  { regs = Array.create 8 0l;
    arrays = [| array0 |];
    flist = 1;
    pc = 0 }

let fetch_instr s =
  let code = s.arrays.(0).(0) in
  let i = Block.get code s.pc in
(*  if s.pc < Array.length code then*) s.pc <- succ s.pc;
  i

let dump ppf s =
  Format.fprintf ppf "pc=%08x " s.pc;
  for i = 0 to 7 do
    Format.fprintf ppf "r%i=%08lx " i s.regs.(i)
  done


let load s b c =
  if b <> 0l then s.arrays.(0).(0) <- Block.copy (get_array s b);
  s.pc <- Int32.to_int c


let save_state fn s =
  let oc = open_out fn in
  output_value oc s;
  close_out oc

let load_state fn =
  let ic = open_in fn in
  let s = input_value ic in
  close_in ic;
  s

let id = "Array2FL"


let gc _ = ()
