exception Found of int list option
exception Error

type state = string

let dest s i = Char.code s.[i lsl 1]
let tick s i = Char.code s.[i lsl 1 + 1]

let seen = Hashtbl.create 1024

let load fn =
  let rex = Pcre.regexp ~study:true "^(\\d+) -> \\((\\d+),(\\d+)\\)$" in
  let ic = if fn = "" then stdin else open_in fn in
  let res = ref [] in
  try
    while true do
      let l = input_line ic in
      try
        let g x = int_of_string (Pcre.get_substring (Pcre.exec ~rex l) x) in
        res := (g 2,g 3) :: !res;
      with
      | Not_found -> ()
    done;
    assert false
  with
  | End_of_file ->
      close_in ic;
      Array.of_list (List.rev !res)

let line n i =
  for j = 0 to n - 2 do
    if j = i then print_string "><"
    else print_char '|'
  done;
  print_newline ();
  flush stdout


let print_pb s =
  let n = String.length s / 2 in
  for i = 0 to n - 1 do
    Printf.printf "(%i,%i,%i)" i (dest s i) (tick s i)
  done;
  print_endline ""

let print_result s res =
  let n = String.length s / 2 in
  print_pb s;
  List.iter (line n) res

let is_done s =
  let n = String.length s / 2 in
  try for i = 0 to n - 1 do 
    if (dest s i != i) || (tick s i != 0) then raise Exit
  done; true
  with Exit -> false

let impossible s =
  let n = String.length s / 2 in
  try for i = 0 to n - 1 do 
    let d = dest s i in
    let cross = ref 0 in
    let t = tick s i in
    for j = 0 to i - 1 do if dest s j > d then incr cross;
    done;
    if (!cross > t - d + i) then raise Exit;
  done; false
  with Exit -> true


let rec swap s i =
  let j = i lsl 1 in
  let d = Char.code s.[j] and t = Char.code s.[j+1] in
  let d' = Char.code s.[j+2] and t'= Char.code s.[j+3] in
  if (d' - i > t') then (print_char 'C'; None)
  else
    let s' = String.copy s in
    s'.[j] <- Char.chr d';
    s'.[j+1] <- Char.chr t';
    s'.[j+2] <- Char.chr d;
    s'.[j+3] <- Char.chr (t-1);
    if (d = i + 1) && (t = 1) then 
      try
	print_char '!'; print_int (String.length s); print_char ':'; print_int i;
	let sub1 = String.sub s' 0 (j + 2) in
	let len1 = String.length sub1 / 2 in
	for k = 0 to len1 - 1 do
	  let j = Char.code sub1.[k lsl 1] in
	  if j >= len1 then raise Error;
	done;
	let sub2 = String.sub s' (j + 4) (String.length s - j - 4) in 
	let len2 = String.length sub2 / 2 in
	let delta = i + 2 in
	for k = 0 to len2 - 1 do
	  let j = Char.code sub2.[k lsl 1] - delta in
	  if j < 0 then raise Error;
	  sub2.[k lsl 1] <- Char.chr (Char.code sub2.[k lsl 1] - delta);
	done;
	(try match Hashtbl.find seen sub2 with None -> raise Error | _ -> ()
	 with Not_found -> ());
	match loop (-1) sub1 with None -> None | Some res1 ->
	  match loop (-1) sub2 with None -> None | Some res2 ->
	    let res2 = List.map (fun x -> x + delta) res2 in
	    Some (res1 @ res2)
      with Error -> None
    else loop i s'

and loop last s =
  print_char '.'; flush stdout;
  let last = -1 in
  if is_done s then (print_char 'A'; Some [])
  else if impossible s then (print_char 'B'; None)
  else try 
    let r = Hashtbl.find seen s in
    print_char 'F';
(*    Printf.printf "found %i\n" (String.length s / 2); flush stdout; *)
    r
  with Not_found ->
    try
      let n = String.length s / 2 in
      let ticks = ref [] in

      let tot = ref 0 in
      for i = 0 to n - 1 do
	tot := !tot + tick s i
      done;
      
(*      let mid = 2 * (!tot / n) in *)
(*      Printf.eprintf "tot=%i n=%i mid=%i\n" !tot n mid; flush stderr; *)
      for i = max 0 (last - 1) to n - 2 do
	let t = tick s i in
	if t > 0 then 
	  let score = n - dest s i in 
(* 	  let score = if t >= mid then t - mid else 2 * mid - t in  *)
	  ticks := (t,i,score (*+ abs (i - (n/2))*)) :: !ticks;
      done;
      let ticks = List.sort 
	(fun (t1,i1,s1) (t2,i2,s2) -> 
	   Pervasives.compare s1 s2
(*
	   let c = Pervasives.compare (abs (i1 - mid)) (abs (i2 - mid)) in
	   if c <> 0 then c
	   else - (Pervasives.compare (t1 : int) t2) *)
	)
	!ticks in

(*      let ticks = List.sort (fun (t1,i1) (t2,i2) -> - (Pervasives.compare i1 i2)) !ticks in *)

(*      Printf.printf "[mid=%i]" mid;
      List.iter (fun (t,i) ->
		   Printf.printf "(%i,%i) " t i) ticks; 
      Printf.printf "\n";  
      flush stdout;  *)
(*      print_char 'A';*)
      List.iter (fun (t,i,_) ->
(*		   print_char '?';*)
		   match swap s i with
		     | None -> ()
		     | Some x -> raise (Found (Some (i::x)))) ticks;
(*      print_char 'B'; *)
(*      print_pb s;   *)
      print_char '-';
      if last < 0 then ((*print_char '-'; flush stdout; *)Hashtbl.add seen s None); 
      None
    with Found r -> print_char '+'; (*print_pb s;*) flush stdout; Hashtbl.add seen s r; r


let create_s pb = 
  let n = Array.length pb in
  let s = String.create (n lsl 1) in
  for i = 0 to n - 1 do 
    let (d,t) = pb.(i) in
    s.[i lsl 1] <- Char.chr d;
    s.[i lsl 1 + 1] <- Char.chr t;
  done;
  s

let main pb =
  let n = Array.length pb in
  let s = create_s pb in
  print_pb s;
  match loop (-1) s with
    | None -> print_endline "No solution"
    | Some x -> print_string "Solution:\n"; List.iter (line n) x

let () = 
  main (load (if Array.length Sys.argv = 1 then "" else Sys.argv.(1)))  
