open Printf

let do_print = true

let countX = ref 0

let trace_X n i =
  incr countX;
  if do_print then (
    for j = 0 to n - 2 do
      if j = i then print_string "><"
      else print_char '|'
    done; 
    print_newline () 
  ) 



exception AgainWithCost

exception Odd

type state = int array

type op = 
  | TrapL of int * int * int
  | TrapR of int * int * int

type ops = op list

exception Found of ops

let transl_op d = function
  | TrapL (i,j,n) -> TrapL (i+d,j+d,n)
  | TrapR (i,j,n) -> TrapR (i+d,j+d,n)

let transl_ops d = List.map (transl_op d)


let print state =
  for i = 0 to Array.length state - 1 do
    printf "%i " state.(i)
  done;
  printf "\n";
  flush stdout

let print_op = function
  | TrapL (i,j,n) -> printf "L(%i,%i,%i)\n" i j n
  | TrapR (i,j,n) -> printf "R(%i,%i,%i)\n" i j n

let print_ops = List.iter print_op

let check state =
  (* Every connex component must have a pair number of ticks *)
  let s = ref 0 in
  let l = ref [] in
  let f () = 
    if !s mod 2 = 1 then raise Odd;
    match !l with
      | [x] -> raise Odd
      | [x;y] when x != y -> raise Odd
      | _ -> s := 0; l := []
  in
  for i = 0 to Array.length state - 1 do
    assert(state.(i) >= 0);
    if state.(i) = 0 then f ()
    else (s := !s + state.(i); l := state.(i) :: !l)
  done;
  f ()

let total state = 
  Array.fold_left (+) 0 state


let trace_op w = function
  | TrapL (i,j,n) ->
      for x = i to j - 1 do trace_X w x done;
      for x = 1 to (n - 1) * 2 do trace_X w (j - 1) done;
      for x = j - 1 downto i do trace_X w x done
  | TrapR (i,j,n) ->
      for x = j - 1 downto i do trace_X w x done;
      for x = 1 to (n - 1) * 2 do trace_X w i done;
      for x = i to j - 1 do trace_X w x done

let trace_ops w ops = List.iter (trace_op w) ops

let apply state = function
  | TrapL (i,j,n) ->
      let k = j - i - 1 in
      assert(i < j);
      assert (n >= 0);
      state.(i) <- state.(i) - n - k;
      state.(j) <- state.(j) - n;
      for x = i + 1 to j - 1 do state.(x) <- state.(x) - 1 done
  | TrapR (i,j,n) ->
      let k = j - i - 1 in
      assert(i < j);
      assert (n >= 0);
      state.(i) <- state.(i) - n;
      state.(j) <- state.(j) - n - k;
      for x = i + 1 to j - 1 do state.(x) <- state.(x) - 1 done
  
let iter_trapeze state i j f =
  let tryop op = 
    let s = Array.copy state in
    apply s op;
    try check s; f op s
    with Odd -> ()
  in

  let a = state.(i) and b = state.(j) in
  let k = j - i - 1 in
  for n = 1 to min b (a - k) do tryop (TrapL (i,j,n)) done;
  for n = 1 to min a (b - k) do tryop (TrapR (i,j,n)) done


let rec components s i =
  if i = Array.length s then []
  else if s.(i) = 0 then components s (succ i)
  else components_in s i (succ i)
and components_in s i j =
  if j = Array.length s then [(i,j-1)]
  else if s.(j) = 0 then (i,j-1) :: components s (succ j)
  else components_in s i (succ j)

let prepare pb =
  let n = Array.length pb in
  for j = n - 1 downto 1 do
    let rec look i = if (fst pb.(i)) = j then i else look (succ i) in
    let i = look 0 in
    for x = i to j - 1 do
      trace_X n x;
      let (d,t) = pb.(x) and (d',t') = pb.(x+1) in
      pb.(x) <- (d',t');
      pb.(x+1) <- (d,t-1)
    done;
  done;

(*
  for i = 0 to n - 1 do
    let (d,t) = pb.(i) in
    printf "(%i,%i)" d t
  done;
  printf "\n";
*)

  Array.map snd pb (* Ticks which remain to do *)

type memo = Success of ops | Failure | Cost of int
let seen = Hashtbl.create 1024

let cost state =  (* lower cost -> better choice *)
  try
    match Hashtbl.find seen state with
      | Success _ -> 0
      | Failure -> max_int
      | Cost i -> i
  with Not_found -> match components state 0 with 
    | [] -> 
	0
    | _::_::_ -> 
	1 (* ??? *)
    | [(i0,j0)] ->
	let mint = ref max_int in
	for i = i0 to j0 - 1 do
	  for j = i + 1 to j0 do
	    iter_trapeze state i j 
	      (fun op s -> 
		 let t = total s in
		 if t < !mint then mint := t)
	  done;
	done;
	Hashtbl.add seen state (Cost !mint);
	!mint

let all_traps state i0 j0 = 
(*  print_char '('; flush stdout; *)
  let traps = ref [] in
  (try
     let n = ref 0 in
  for i = i0 to j0 - 1 do
    for j = i + 1 to j0 do
      iter_trapeze state i j 
	(fun op s -> traps := (op,s) :: !traps; incr n; 
(*	   if !n = 50000 then raise Exit *)
	)
    done;
  done;
  with Exit -> ());
(*  print_char ')'; flush stdout;  *)
  !traps
 
let sort_traps cost traps =
  List.sort (fun (t1,_,_) (t2,_,_) -> compare t1 t2)
    (List.map (fun (op,s) -> cost s, op, s) traps)


let rec loop state = 
  let r = 
    try match Hashtbl.find seen state with Cost c -> raise Not_found | r -> r
    with Not_found ->
      let r = realloop state in
      Hashtbl.add seen state r;
      r
  in
(*  if r = Failure then print_char '-' else print_char '+'; *)
  r
      
and split state comps = List.fold_left
  (fun accu (i,j) ->
     let l = j - i + 1 in
     let a = Array.create l 0 in
     Array.blit state i a 0 l;
     match loop a with Failure -> raise Exit | Success l ->
       transl_ops i l @ accu | Cost _ -> assert false
  ) [] comps
  
and realloop state : memo = match components state 0 with
  | [] -> 
      Success []
  | _::_::_ as c -> 
(*      print_char 'S'; flush stdout; *)
      (try Success (split state c) with Exit -> Failure)
  | [(i0,j0)] ->
      let atraps = all_traps state i0 j0 in

      try
      let traps = sort_traps total atraps in
      let n = ref 0 in
(*      let l = List.length traps in
(*      printf "{%i}" (total state); flush stdout;  *)
      let t = total state in *)
      try 
	List.iter (fun (_,op,s) -> 
(*		     printf "[%i:%i/%i]" t !n l; flush stdout; *)
		     if (!n = 10) then raise AgainWithCost;
		     incr n;
		     match loop s with
		       | Success r -> raise (Found (op::r))
		       | _ -> ()) traps;
	Failure
      with Found r -> Success r

      with AgainWithCost ->
      let traps = sort_traps cost atraps in
      let n = ref 0 in
(*      let l = List.length traps in
      let t = total state in *)
      try 
	List.iter (fun (_,op,s) -> 
(*		     printf "{%i:%i/%i}" t !n l; flush stdout; *)
		     if (!n = 5) then raise Exit;
		     incr n;
		     match loop s with
		       | Success r -> raise (Found (op::r))
		       | _ -> ()) traps;
	Failure
      with Found r -> Success r | Exit -> Failure

let load fn =
  let rex = Pcre.regexp ~study:true "^(\\d+) -> \\((\\d+),(\\d+)\\)$" in
  let ic = if fn = "" then stdin else open_in fn in
  let res = ref [] in
  try
    while true do
      let l = input_line ic in
      try
        let g x = int_of_string (Pcre.get_substring (Pcre.exec ~rex l) x) in
        res := (g 2,g 3) :: !res;
      with
      | Not_found -> ()
    done;
    assert false
  with
  | End_of_file ->
      close_in ic;
      Array.of_list (List.rev !res)

let reduce2 d s =
  let n = Array.length s in
  let kmax = ref 0 in
  let h = Array.init (n / 2) 
    (fun x ->
       let i = x * 2 in
       let a = s.(i) and b = s.(i+1) in
       let k = max 0 (min (a - d) (b - d)) in
       s.(i) <- a - k; s.(i+1) <- b - k;
       kmax := max !kmax k;
       max k 0
    ) in
  for k = 1 to !kmax do
    for r = 1 to 2 do
      incr countX;
      for i = 0 to n / 2 - 1 do
	if h.(i) >= k then print_string "><"
	else print_string "||";
      done;
      print_newline ()
    done;
  done

let () = 
  print_string "bbarker\nplinko\nverify\n";
  let pb = load (if Array.length Sys.argv = 1 then "" else Sys.argv.(1)) in

  let state = prepare pb in
  Printf.eprintf "prepare: %i lines\n" !countX; countX := 0;
  Printf.eprintf "before reduce: %i remain\n" (total state); 
  reduce2 5 state;
  Printf.eprintf "reduce: %i lines\n" !countX; countX := 0;
  Printf.eprintf "after reduce: %i remain\n" (total state); 
  flush stderr;
  
  match loop state with
  | Success r -> 
      trace_ops (Array.length state) r; print_endline "\nalain\nalain\n";
      Printf.eprintf "loop: %i lines\n" !countX; countX := 0;
  | _ -> print_endline "No solution"
